﻿using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;
using System.Linq.Expressions;
using System.Linq;
using System.Reflection;

namespace PropertyChangedDemo
{
    public class BaseModel : INotifyPropertyChanged
    {
        #region INotifyPropertyChanged Members

        public event PropertyChangedEventHandler PropertyChanged;

        protected virtual void RaiseOnPropertyChanged(string propertyName)
        {
            var handler = PropertyChanged;
            if (handler != null)
                handler(this, new PropertyChangedEventArgs(propertyName));
        }

        #endregion


        protected static DefineChangeObject<TChanging, TListening> For<TChanging, TListening>(Expression<Func<TListening, object>> propertyLambdaExpression)
        {
            return new DefineChangeObject<TChanging, TListening>(new BasicPropertyInfo(ExpressionHelper.GetMemberInfo(propertyLambdaExpression)));
        }

        protected static DefineChangeObject<TListening> For<TListening>(Expression<Func<TListening, object>> propertyLambdaExpression)
        {
            return new DefineChangeObject<TListening>(new BasicPropertyInfo(ExpressionHelper.GetMemberInfo(propertyLambdaExpression)));
        }


        #region DefineChangeObject<TListener>

        public class DefineChangeObject<TListener>
        {
            public BasicPropertyInfo ChangingProperty { get; set; }

            public DefineChangeObject(BasicPropertyInfo changingProperty)
            {
                this.ChangingProperty = changingProperty;
            }

            public BasicPropertyInfo Define(Expression<Func<TListener, object>> valueExpression)
            {
                var valueMembers = ExpressionHelper.GetSourceProperties(valueExpression);

                var listeningMembers = valueMembers.Where(x => x.ReflectedType.IsAssignableFrom(typeof(TListener))).Select(x => x.Name).ToArray();

                //If we can't detect any properties on the changing object, then just store that action against the empty string.
                //This'll be run when all actions are run.
                if (!listeningMembers.Any())
                    listeningMembers = new string[] { "" };

                Action<TListener> a = (listener) =>
                    ((PropertyInfo)this.ChangingProperty.Member).SetValue(listener, valueExpression.Compile().Invoke(listener), null);

                PropertyChangedActionManager<TListener>.Instance.RegisterListenerChangedAction(a, listeningMembers);
                return this.ChangingProperty;
            }

        }

        #endregion


        #region DefineChangeObject<TSource, TListener>

        public class DefineChangeObject<TSource, TListener>
        {
            //public string[] Sources { get; private set; }
            public BasicPropertyInfo ChangingProperty { get; set; }

            public DefineChangeObject(BasicPropertyInfo changingProperty)
            {
                this.ChangingProperty = changingProperty;
            }

            public BasicPropertyInfo Define(Expression<Func<TSource, TListener, object>> valueExpression)
            {
                var valueMembers = ExpressionHelper.GetSourceProperties(valueExpression);

                var changingMembers = valueMembers.Where(x => x.ReflectedType.IsAssignableFrom(typeof(TSource))).Select(x => x.Name).ToArray();
                var listeningMembers = valueMembers.Where(x => x.ReflectedType.IsAssignableFrom(typeof(TListener))).Select(x => x.Name).ToArray();

                //If we can't detect any properties on the changing object, then just store that action against the empty string.
                //This'll be run when all actions are run.
                if (!changingMembers.Any())
                    changingMembers = new string[] { "" };

                Action<TSource, TListener> a = (changing, listening) =>
                    ((PropertyInfo)this.ChangingProperty.Member).SetValue(listening, valueExpression.Compile().Invoke(changing, listening), null);

                PropertyChangedActionManager<TSource, TListener>.Instance.RegisterSourceChangedAction(a, changingMembers);
                PropertyChangedActionManager<TSource, TListener>.Instance.RegisterListenerChangedAction(a, listeningMembers);
                return this.ChangingProperty;
            }
        }

        #endregion
    }
}
